import os

import pytest

from testcontainers.core.container import DockerContainer
from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import wait_for_logs
from testcontainers.minio import MinioContainer
from testcontainers.nats import NatsContainer
from testcontainers.postgres import PostgresContainer


@pytest.fixture(scope="session")
def local_environment(network, minio, nats, postgres):
    with Network() as network:
        postgres = PostgresContainer(
            "postgres:13.3-alpine",
            network=network.name,
            hostname="postgres",
            username="postgres",
            password="postgres",
            dbname="postgres",
        )
        minio = MinioContainer("minio/minio:latest", network=network.name, hostname="minio")
        nats = NatsContainer("nats:2.10-alpine", network=network.name, hostname="nats")
        nats = nats.with_command("-js")
        with postgres as postgres, minio as minio, nats as nats:
            foreverbull = DockerContainer(os.environ.get("BROKER_IMAGE", ""))
            foreverbull.with_network(network)
            foreverbull.with_env("POSTGRES_URL", "postgres://postgres:postgres@postgres:5432/postgres")
            foreverbull.with_env("NATS_URL", "nats://nats:4222")
            foreverbull.with_env("MINIO_URL", "minio:9000")
            foreverbull.with_env("BACKTEST_IMAGE", os.environ.get("BACKTEST_IMAGE", "lhjnilsson/zipline:latest"))
            foreverbull.with_volume_mapping("/var/run/docker.sock", "/var/run/docker.sock", mode="rw")
            with foreverbull as foreverbull:
                wait_for_logs(foreverbull, "RUNNING", 10)
                yield foreverbull



import os

from datetime import date
from datetime import datetime

import pytest

from docker.models.containers import Container
from sqlalchemy import create_engine
from testcontainers.core.config import testcontainers_config
from testcontainers.postgres import PostgresContainer

from foreverbull.pb import pb_utils
from foreverbull.pb.foreverbull.backtest import backtest_pb2
from foreverbull_testing.data import AssetManager
from foreverbull_testing.data import PortfolioManager

from . import database


@pytest.fixture(scope="session")
def fb_database():
    postgres = PostgresContainer("postgres:alpine")
    postgres = postgres.with_volume_mapping("postgres_data", "/var/lib/postgresql/data", mode="rw")
    postgres = postgres.with_name("fbull_client_testing_pg")
    client = postgres.get_docker_client()
    containers = client.client.api.containers(filters={"name": "fbull_client_testing_pg"})
    if containers:
        postgres._container = Container(attrs=containers[0])
        engine = create_engine(postgres.get_connection_url())
    else:
        testcontainers_config.ryuk_disabled = True
        postgres.start()
        engine = create_engine(postgres.get_connection_url())
        database.Base.metadata.create_all(engine)

    os.environ["DATABASE_URL"] = postgres.get_connection_url()

    def verify_or_populate(
        entity: backtest_pb2.Backtest | None = None,
        start: date | None = None,
        end: datetime | None = None,
        symbols: list[str] | None = None,
    ):
        if entity is None and start is None and end is None and symbols is None:
            raise ValueError("At least one of entity or start, end, and symbols must be provided")
        if not entity:
            if type(start) is str:
                start = datetime.strptime(start, "%Y-%m-%d").date()
            if type(end) is str:
                end = datetime.strptime(end, "%Y-%m-%d").date()

            assert start and end and symbols
            entity = backtest_pb2.Backtest(
                start_date=pb_utils.from_pydate_to_proto_date(start),
                end_date=pb_utils.from_pydate_to_proto_date(end),
                symbols=symbols,
            )
        if not database.verify(engine, entity):
            database.populate(engine, entity)

    yield engine, verify_or_populate


@pytest.fixture
def asset_manager(fb_database):
    engine, vop = fb_database
    with engine.connect() as conn:
        yield AssetManager(conn, vop)


@pytest.fixture
def portfolio_manager(fb_database):
    engine, _ = fb_database
    with engine.connect() as conn:
        yield PortfolioManager(conn)



from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Union

import pandas

from pandas import DataFrame
from pandas import read_sql_query
from sqlalchemy import Connection

from foreverbull import Asset  # type: ignore
from foreverbull import Assets  # type: ignore
from foreverbull import Portfolio
from foreverbull.pb import pb_utils
from foreverbull.pb.foreverbull.finance import finance_pb2


@dataclass
class Position:
    symbol: str
    amount: int


class Asset(Asset):
    def __init__(
        self, db: Connection, start: str | datetime, end: str | datetime, symbol: str, metrics: dict[str, Any] = {}
    ):
        self._start = start
        self._end = end
        self._symbol = symbol
        self._stock_data = read_sql_query(
            f"""Select symbol, time, high, low, open, close, volume
            FROM ohlc WHERE time BETWEEN '{start}' AND '{end}'
            AND symbol='{symbol}'""",
            db,
        )
        self.metrics = metrics

    def get_metric[T: (int, float, bool, str)](self, key: str) -> Union[T, None]:
        try:
            return self.metrics[key]
        except KeyError:
            return None

    def set_metric[T: (int, float, bool, str)](self, key: str, value: T) -> None:
        self.metrics[key] = value

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def stock_data(self) -> DataFrame:
        return self._stock_data


class Assets(Assets):
    def __init__(
        self,
        db: Connection,
        start: str | datetime,
        end: str | datetime,
        symbols: list[str],
        metrics: dict[str, Any] = {},
    ):
        self._db = db
        self._start = start
        self._end = end
        self._symbols = symbols
        self._stock_data = read_sql_query(
            f"""Select symbol, time, high, low, open, close, volume
            FROM ohlc WHERE time BETWEEN '{start}' AND '{end}' AND symbol IN {tuple(symbols) if len(symbols) > 1 else f"('{symbols[0]}')" }""",
            db,
        )
        self._stock_data.set_index(["symbol", "time"], inplace=True)
        self._stock_data.sort_index(inplace=True)
        self.metrics = metrics

    def get_metrics[T: (int, float, bool, str)](self, key: str) -> pandas.Series:
        try:
            return pandas.Series(self.metrics[key])
        except KeyError:
            return pandas.Series()

    def set_metrics[T: (int, float, bool, str)](self, key: str, value: dict[str, T]) -> None:
        self.metrics[key] = value

    @property
    def symbols(self) -> list[str]:
        return self._symbols

    def __iter__(self):
        for symbol in self._symbols:
            yield Asset(self._db, self._start, self._end, symbol)

    @property
    def stock_data(self) -> DataFrame:
        return self._stock_data


class AssetManager:
    def __init__(self, db: Connection, vop):
        self._db = db
        self._vop = vop

    def get_asset(self, start: str | datetime, end: str | datetime, symbol: str) -> Asset:
        self._vop(None, start, end, [symbol])
        return Asset(self._db, start, end, symbol)

    def get_assets(
        self, start: str | datetime, end: str | datetime, symbols: list[str], metrics: dict[str, Any] = {}
    ) -> Assets:
        self._vop(None, start, end, symbols)
        return Assets(self._db, start, end, symbols, metrics)


class PortfolioManager:
    def __init__(self, db: Connection):
        self._db = db

    def get_portfolio(
        self, dt: datetime | None = None, portfolio_value: float = 100_000, positions: list[Position] = []
    ) -> Portfolio:
        proto_positions = []
        for position in positions:
            proto_positions.append(
                finance_pb2.Position(
                    symbol=position.symbol,
                    amount=position.amount,
                )
            )
        return Portfolio(
            finance_pb2.Portfolio(
                timestamp=pb_utils.to_proto_timestamp(dt) if dt else None,
                positions=proto_positions,
                portfolio_value=portfolio_value,
            ),
            self._db,
        )
